# -*- coding: utf-8 -*-
"""
Created on 19 November 2017
@author: Charlie Lewis
"""
import logging
from collections import defaultdict

from poseidon_core.controllers.faucet.acls import ACLs
from poseidon_core.controllers.faucet.config import FaucetLocalConfGetSetter
from poseidon_core.controllers.faucet.config import FaucetRemoteConfGetSetter
from poseidon_core.controllers.faucet.helpers import parse_rules
from poseidon_core.volos.acls import Acl


class Parser:

    def __init__(self,
                 mirror_ports=None,
                 proxy_mirror_ports=None,
                 reinvestigation_frequency=None,
                 max_concurrent_reinvestigations=None,
                 ignore_vlans=None,
                 ignore_ports=None,
                 tunnel_vlan=None,
                 tunnel_name=None,
                 faucetconfrpc_address=None,
                 faucetconfrpc_client=None,
                 copro_port=None,
                 copro_vlan=None,
                 faucetconfgetsetter_cl=FaucetRemoteConfGetSetter):
        self.logger = logging.getLogger('parser')
        self.mirror_ports = mirror_ports
        self.proxy_mirror_ports = proxy_mirror_ports
        self.reinvestigation_frequency = reinvestigation_frequency
        self.max_concurrent_reinvestigations = max_concurrent_reinvestigations
        self.ignore_vlans = ignore_vlans
        self.ignore_ports = ignore_ports
        self.copro_port = copro_port
        self.copro_vlan = copro_vlan
        self.tunnel_vlan = tunnel_vlan
        self.tunnel_name = tunnel_name
        if faucetconfrpc_address is None:
            faucetconfgetsetter_cl = FaucetLocalConfGetSetter
        if faucetconfrpc_address:
            server = faucetconfrpc_address.split(':')[0]
        else:
            server = ''
        self.faucetconfgetsetter = faucetconfgetsetter_cl(
            client_key='/certs/%s.key' % faucetconfrpc_client,
            client_cert='/certs/%s.crt' % faucetconfrpc_client,
            ca_cert='/certs/%s-ca.crt' % server,
            server_addr=faucetconfrpc_address)
        self.mac_table = {}
        self.mirror_counts = defaultdict(int)
        self._set_default_switch_conf()

    def _read_faucet_conf(self):
        return self.faucetconfgetsetter.read_faucet_conf(config_file=None)

    def _write_faucet_conf(self, faucet_conf=None):
        return self.faucetconfgetsetter.write_faucet_conf(config_file=None, faucet_conf=faucet_conf)

    def _get_dps(self):
        return self.faucetconfgetsetter.get_dps()

    def _get_switch_conf(self, dp):
        return self.faucetconfgetsetter.get_switch_conf(dp)

    def _get_port_conf(self, dp, port):
        return self.faucetconfgetsetter.get_port_conf(dp, port)

    def _get_stack_root_switch(self):
        return self.faucetconfgetsetter.get_stack_root_switch()

    def _get_acls(self):
        faucet_conf = self._read_faucet_conf()
        acls = Acl(faucetconfgetsetter=self.faucetconfgetsetter)
        acls.read(config_yaml=faucet_conf)
        return acls

    def _set_acls(self, acls):
        self.faucetconfgetsetter.set_acls(acls)

    def _update_switch_conf(self, dp, switch_conf):
        self.faucetconfgetsetter.update_switch_conf(dp, switch_conf)

    def _set_port_conf(self, dp, port, port_conf):
        self.faucetconfgetsetter.set_port_conf(dp, port, port_conf)

    def _clear_mirror_port(self, switch, port):
        self.faucetconfgetsetter.clear_mirror_port(switch, port)

    def _mirror_port(self, switch, mirror_port, port):
        self.faucetconfgetsetter.mirror_port(switch, mirror_port, port)

    def _unmirror_port(self, switch, mirror_port, port):
        self.faucetconfgetsetter.unmirror_port(switch, mirror_port, port)

    def _set_default_switch_conf(self):
        # TODO: make smarter with more complex configs (backup original values, etc)
        if self.mirror_ports:
            root_stack_switch = self._get_stack_root_switch()
            root_mirror_port = None
            if root_stack_switch:
                acls = self._get_acls()
                root_mirror_port = self.mirror_ports.get(
                    root_stack_switch, None)
                if self.tunnel_name in acls.acls:
                    del acls.acls[self.tunnel_name]
                # Safety rule to prevent looped packets being output to the loop.
                acls.add_rule(
                    self.tunnel_name,
                    {'rule': {'vlan_vid': self.tunnel_vlan, 'actions': {'allow': 0}}})
                # Tunnel back to root.
                acls.add_rule(
                    self.tunnel_name,
                    {'rule': {'actions': {'allow': 0, 'output': {'tunnel': {
                        'type': 'vlan', 'tunnel_id': self.tunnel_vlan,
                        'dp': root_stack_switch, 'port': root_mirror_port}}}}})
                # Merge ACL updates, if any back in
                self._set_acls(acls.acls)

            for switch, mirror_port in self.mirror_ports.items():
                if not self._get_switch_conf(switch):
                    continue
                if self.reinvestigation_frequency:
                    switch_conf = {
                        'timeout': (self.reinvestigation_frequency * 2) + 1,
                        'arp_neighbor_timeout': self.reinvestigation_frequency}
                    self._update_switch_conf(switch, switch_conf)
                # If stacking was detected, provision tunnel config on non root switches.
                # Poseidon's mirror NIC must be connected to the root switch's mirror port.
                # Non mirror switches must have a loopback plug installed in their mirror port.
                if root_stack_switch and root_mirror_port:
                    mirror_port_conf = self._get_port_conf(switch, mirror_port)
                    if not mirror_port_conf:
                        continue
                    for existing_key in set(mirror_port_conf.keys()):
                        if existing_key not in ('mirror',):
                            del mirror_port_conf[existing_key]
                    if root_stack_switch == switch:
                        mirror_port_conf.update({
                            'description': 'Poseidon local mirror',
                            'output_only': True,
                        })
                    else:
                        mirror_port_conf.update({
                            'description': 'Poseidon remote mirror (loopback plug)',
                            'acls_in': [self.tunnel_name],
                            'coprocessor': {'strategy': 'vlan_vid'},
                        })
                    self._set_port_conf(switch, mirror_port, mirror_port_conf)

    def proxy_mirror_port(self, switch, port):
        self.logger.debug('checking for proxy ports')
        if self.proxy_mirror_ports and self._get_switch_conf(switch):
            proxy_ports = self.proxy_mirror_ports.get(switch, None)
            if proxy_ports:
                self.logger.debug(f'found proxy port: {switch} port {port}')
                switch, port = proxy_ports
        return switch, port

    def mirror_switch_port(self, switch):
        if self.mirror_ports:
            switch_mirror_port = self.mirror_ports.get(switch, None)
            if switch_mirror_port:
                if self._get_port_conf(switch, switch_mirror_port):
                    return switch_mirror_port
        self.logger.warning('no mirror port for switch %s' % switch)
        return None

    def clear_mirrors(self):
        dps = self._get_dps()
        if dps:
            for switch in dps:
                mirror_port = self.mirror_switch_port(switch)
                if mirror_port:
                    self._clear_mirror_port(switch, mirror_port)

    def config_mirror(self, action, switch, port):
        switch, port = self.proxy_mirror_port(switch, port)
        mirror_port = self.mirror_switch_port(switch)
        if mirror_port:
            mirror_key = (switch, port)
            if action in ('mirror', 'unmirror'):
                self.logger.info(f'request {action} of {mirror_key}')
                if action == 'mirror':
                    self._mirror_port(switch, mirror_port, port)
                    self.mirror_counts[mirror_key] += 1
                elif action == 'unmirror':
                    if self.mirror_counts[mirror_key]:
                        self.mirror_counts[mirror_key] -= 1
                        if not self.mirror_counts[mirror_key]:
                            self.logger.info(
                                f'removing last remaining mirror on {mirror_key}')
                            self._unmirror_port(switch, mirror_port, port)
                count = self.mirror_counts[mirror_key]
                self.logger.info(f'mirroring {count} MACs on {mirror_key}')
            else:
                self.logger.error(f'Unknown mirror action {action} for {mirror_key}')
        else:
            self.logger.error(
                f'Unable to configure mirror/unmirror {switch}:{port} due to warnings')

    def config_acls(self, rules_file, endpoints, force_apply_rules, force_remove_rules, coprocess_rules_files):
        rules_doc = parse_rules(rules_file)
        self._read_faucet_conf()
        self.faucetconfgetsetter.faucet_conf = ACLs().apply_acls(
            rules_file, endpoints,
            force_apply_rules, force_remove_rules,
            coprocess_rules_files, self.faucetconfgetsetter.faucet_conf, rules_doc)
        self._write_faucet_conf()

    def config(self, action, port, switch, rules_file=None,
               endpoints=None, force_apply_rules=None, force_remove_rules=None,
               coprocess_rules_files=None):
        if action in ('shutdown', 'apply_routes'):
            # TODO: not implemented.
            pass
        elif action in ('mirror', 'unmirror'):
            self.config_mirror(action, switch, port)
        elif action == 'apply_acls':
            self.config_acls(
                rules_file, endpoints, force_apply_rules, force_remove_rules, coprocess_rules_files)
        else:
            self.logger.warning('Unknown action: {0}'.format(action))

    def ignore_event(self, message):
        for message_type in ('L2_LEARN',):
            message_body = message.get(message_type, None)
            if message_body:
                switch = str(message['dp_name'])
                port_no = message_body.get('port_no', None)
                vlan = message_body.get('vid', None)
                # When stacking is in use, we only want to learn on switch, that a host is local to.
                if 'stack_descr' in message_body:
                    self.logger.debug(
                        'ignoring event because learning from a stack port')
                    return True
                if self.ignore_vlans:
                    if vlan in self.ignore_vlans:
                        self.logger.debug(
                            'ignoring event because VLAN %s ignored' % vlan)
                        return True
                if self.ignore_ports:
                    for ignore_switch, ignore_port_no in self.ignore_ports.items():
                        if ignore_switch == switch and ignore_port_no == port_no:
                            self.logger.debug(
                                'ignoring event because switch %s port %s is ignored' % (
                                    switch, port_no))
                            return True
                if self.proxy_mirror_ports:
                    for s in self.proxy_mirror_ports:
                        if (switch == self.proxy_mirror_ports[s][0] and
                            port_no == self.proxy_mirror_ports[s][1]):
                            self.logger.debug(
                                'ignoring event because switch %s port %s is being a proxy' % (
                                    switch, port_no))
                            return True
                # Not on any ignore list, don't ignore.
                return False
        # Not a message we are interested in, ignore it.
        return True

    def event(self, message):
        dp_name = str(message['dp_name'])

        if 'L2_LEARN' in message:
            self.logger.debug(
                'got faucet message for l2_learn: {0}'.format(message))
            message = message['L2_LEARN']
            eth_src = message['eth_src']
            vlan_str = 'VLAN%s' % message['vid']
            data = {
                'ip-address': message['l3_src_ip'],
                'ip-state': 'L2 learned',
                'mac': eth_src,
                'segment': dp_name,
                'port': str(message['port_no']),
                'vlan': vlan_str,
                'tenant': vlan_str,
                'active': 1}

            if eth_src in self.mac_table:
                if data in self.mac_table[eth_src]:
                    self.mac_table[eth_src].remove(data)
                self.mac_table[eth_src].insert(0, data)
            else:
                self.mac_table[eth_src] = [data]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The main entrypoint for Poseidon

Created on 3 December 2018
@author: Charlie Lewis
"""
import logging

from poseidon_core.helpers.log import Logger
from poseidon_core.helpers.rabbit import Rabbit
from poseidon_core.monitor import Monitor

def main():  # pragma: no cover
    logging.getLogger('pika').setLevel(logging.WARNING)
    Logger()
    logger = logging.getLogger('main')
    pmain = Monitor(logger)
    host = pmain.controller['FA_RABBIT_HOST']
    port = int(pmain.controller['FA_RABBIT_PORT'])

    rabbit = Rabbit()
    exchange = 'topic-poseidon-internal'
    binding_key = ['poseidon.algos.#', 'poseidon.action.#']
    rabbit.make_rabbit_connection(
        host, port, exchange, binding_key)
    rabbit.start_channel(
        pmain.rabbit_callback, pmain.m_queue)
    pmain.rabbits.append(rabbit)

    rabbit = Rabbit()
    exchange = pmain.controller['FA_RABBIT_EXCHANGE']
    binding_key = [pmain.controller['FA_RABBIT_ROUTING_KEY']+'.#']
    rabbit.make_rabbit_connection(
        host, port, exchange, binding_key)
    rabbit.start_channel(
        pmain.rabbit_callback, pmain.m_queue)
    pmain.rabbits.append(rabbit)

    pmain.schedule_thread.start()

    # loop here until told not to
    try:
        pmain.process()
    except Exception as e:
        logger.info('Waiting for connection to RabbitMQ...')



# -*- coding: utf-8 -*-
'''
Created on 17 November 2017
@author: Charlie Lewis
'''
import ipaddress
import logging

from poseidon_core.controllers.faucet.parser import Parser
from poseidon_core.volos.coprocessor import Coprocessor
from poseidon_core.volos.volos import Volos


class FaucetProxy(Parser):

    def __init__(self,
                 controller,
                 *args,
                 **kwargs):
        '''Initializes Faucet object.'''
        self.mirror_ports = controller['MIRROR_PORTS']
        self.proxy_mirror_ports = controller['controller_proxy_mirror_ports']
        self.tunnel_vlan = controller['tunnel_vlan']
        self.tunnel_name = controller['tunnel_name']
        self.learn_pub_adds = controller['LEARN_PUBLIC_ADDRESSES']
        self.reinvestigation_frequency = controller['reinvestigation_frequency']
        self.max_concurrent_reinvestigations = controller['max_concurrent_reinvestigations']
        self.trunk_ports = controller['trunk_ports']
        self.ignore_vlans = controller['ignore_vlans']
        self.ignore_ports = controller['ignore_ports']
        self.faucetconfrpc_address = controller['faucetconfrpc_address']
        self.faucetconfrpc_client = controller['faucetconfrpc_client']

        super(FaucetProxy, self).__init__(
            self.mirror_ports,
            self.proxy_mirror_ports,
            self.reinvestigation_frequency,
            self.max_concurrent_reinvestigations,
            self.ignore_vlans,
            self.ignore_ports,
            self.tunnel_vlan,
            self.tunnel_name,
            faucetconfrpc_address=self.faucetconfrpc_address,
            faucetconfrpc_client=self.faucetconfrpc_client,
            *args, **kwargs)

        # parse volos config
        self.volos = Volos(controller)
        self.coprocessor = Coprocessor(controller)
        self.logger = logging.getLogger('faucet')
        self.mac_table = {}

    @staticmethod
    def format_endpoints(data):
        '''
        return only the information needed for the application
        '''
        ret_list = list()
        for d in data:
            md = d[0]
            d.reverse()
            for i, _ in enumerate(d):
                ipv4_set = False
                ipv6_set = False
                if 'ip-address' in d[i]:
                    if ':' in d[i]['ip-address']:
                        md['ipv6'] = d[i]['ip-address']
                        ipv6_set = True
                    else:
                        md['ipv4'] = d[i]['ip-address']
                        ipv4_set = True
                if 'ipv4' in md:
                    ipv4_set = True
                if 'ipv6' in md:
                    ipv6_set = True
            if not ipv4_set:
                md['ipv4'] = 0
            if not ipv6_set:
                md['ipv6'] = 0
            if 'ip-state' in md:
                del md['ip-state']
            if 'ip-address' in md:
                del md['ip-address']

            md.update({
                'controller_type': 'faucet',
                'controller': '',
                'name': None})
            ret_list.append(md)
        return ret_list

    def get_endpoints(self, messages=None):
        retval = []

        if messages:
            self.logger.debug('faucet messages: {0}'.format(messages))
            for message in messages:
                if not self.ignore_event(message):
                    self.event(message)
        for mac in self.mac_table:
            if self.learn_pub_adds:
                retval.append(self.mac_table[mac])
            else:
                # only allow private addresses
                first_entry = self.mac_table[mac][0]
                if 'ip-address' in first_entry and (
                        first_entry['ip-address'] == 'None' or
                        first_entry['ip-address'] is None or
                        not ipaddress.ip_address(first_entry['ip-address']).is_global):
                    retval.append(self.mac_table[mac])
        return retval

    def update_acls(self, rules_file=None, endpoints=None, force_apply_rules=None, force_remove_rules=None):
        self.logger.debug('updating acls')
        self.config('apply_acls', None, None,
                    rules_file=rules_file, endpoints=endpoints,
                    force_apply_rules=force_apply_rules,
                    force_remove_rules=force_remove_rules)
        # TODO check if config was successfully updated
        return True

    def shutdown_ip(self, ip_addr, shutdown=True, mac_addr=None):
        shutdowns = []
        port = 0
        switch = None
        self.config('shutdown', int(port), switch)
        # TODO check if config was successfully updated
        return shutdowns

    def shutdown_endpoint(self):
        port = 0
        switch = None
        self.config('shutdown', int(port), switch)
        # TODO check if config was successfully updated

    def _mac_switch_port(self, my_mac):
        try:
            entry = self.mac_table[my_mac][0]
            return (entry['segment'], entry['port'])
        except (KeyError, IndexError):
            return (None, None)

    def mirror_mac(self, my_mac, my_switch, my_port):
        self.logger.debug('mirroring mac %s', my_mac)
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            self.config('mirror', int(port), switch)
        return True

    def unmirror_mac(self, my_mac, my_switch, my_port):
        self.logger.debug('unmirroring mac %s', my_mac)
        switch, port = self._mac_switch_port(my_mac)
        if port and switch:
            trunk = False
            for sw in self.trunk_ports:
                if sw == switch and self.trunk_ports[sw].split(',')[1] == str(port):
                    trunk = True
            if not trunk:
                self.config('unmirror', int(port), switch)
            else:
                self.logger.debug('not unmirroring a trunk port')
                return False
        return True

    def coprocess_mac(self, my_mac):
        self.logger.debug('coprocess mac: {0}'.format(my_mac))
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            self.config('coprocess', int(port), switch)
        return True

    def uncoprocess_mac(self, my_mac):
        self.logger.debug('uncoprocess mac: {0}'.format(my_mac))
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            self.config('uncoprocess', int(port), switch)
        return True



import logging
import os
import sys

import yaml
from faucetconfrpc.faucetconfrpc_client_lib import FaucetConfRpcClient

from poseidon_core.controllers.faucet.helpers import get_config_file
from poseidon_core.controllers.faucet.helpers import yaml_in
from poseidon_core.controllers.faucet.helpers import yaml_out


class FaucetConfGetSetter:

    DEFAULT_CONFIG_FILE = ''

    def __init__(self, **_kwargs):
        self.faucet_conf = {}

    def read_faucet_conf(self, config_file):
        # TODO dummy function for set_acls
        return

    def write_faucet_conf(self, config_file):
        # TODO dummy function for set_acls
        return

    @staticmethod
    def config_file_path(config_file):
        return config_file

    def set_acls(self, acls):
        self.read_faucet_conf(config_file=None)
        self.faucet_conf['acls'] = acls
        self.write_faucet_conf(config_file=None)

    def get_dps(self):  # pragma: no cover
        raise NotImplementedError

    def get_switch_conf(self, dp):
        return self.get_dps().get(dp, None)

    def get_port_conf(self, dp, port):
        switch_conf = self.get_switch_conf(dp)
        if not switch_conf:
            return None
        return switch_conf['interfaces'].get(port, None)

    def get_stack_root_switch(self):
        root_stack_switch = [
            switch for switch, switch_conf in self.get_dps().items()
            if switch_conf.get('stack', {}).get('priority', None)]
        if root_stack_switch:
            return root_stack_switch[0]
        return None

    def set_port_conf(self, dp, port, port_conf):  # pragma: no cover
        raise NotImplementedError

    def update_switch_conf(self, dp, switch_conf):  # pragma: no cover
        raise NotImplementedError

    def mirror_port(self, dp, mirror_port, port):  # pragma: no cover
        raise NotImplementedError

    def unmirror_port(self, dp, mirror_port, port):  # pragma: no cover
        raise NotImplementedError

    def clear_mirror_port(self, dp, mirror_port):  # pragma: no cover
        raise NotImplementedError


class FaucetLocalConfGetSetter(FaucetConfGetSetter):

    def read_faucet_conf(self, config_file):
        if not config_file:
            config_file = self.DEFAULT_CONFIG_FILE
        config_file = get_config_file(config_file)
        faucet_conf = yaml_in(config_file)
        if faucet_conf is None:
            logging.error('Faucet config is empty, exiting.')
            sys.exit(1)
        if isinstance(faucet_conf, dict):
            self.faucet_conf = faucet_conf
        return self.faucet_conf

    def write_faucet_conf(self, config_file=None, faucet_conf=None):
        if not config_file:
            config_file = self.DEFAULT_CONFIG_FILE
        if faucet_conf is None:
            faucet_conf = self.faucet_conf
        self.faucet_conf = faucet_conf
        config_file = get_config_file(config_file)
        return yaml_out(config_file, self.faucet_conf)

    def get_dps(self):
        self.read_faucet_conf(config_file=None)
        return self.faucet_conf.get('dps', {})

    def set_port_conf(self, dp, port, port_conf):
        switch_conf = self.get_switch_conf(dp)
        if not switch_conf:
            return None
        switch_conf['interfaces'][port] = port_conf
        self.write_faucet_conf()

    def update_switch_conf(self, dp, switch_conf):
        self.faucet_conf['dps'][dp].update(switch_conf)
        self.write_faucet_conf()

    def _get_mirrored_ports(self, dp, mirror_port):
        mirror_interface_conf = self.get_port_conf(dp, mirror_port)
        mirrored_ports = None
        if mirror_interface_conf:
            mirrored_ports = mirror_interface_conf.get('mirror', None)
        return mirror_interface_conf, mirrored_ports

    def _set_mirror_config(self, dp, mirror_port, mirror_interface_conf, ports=None):
        if ports:
            if isinstance(ports, set):
                ports = list(ports)
            if not isinstance(ports, list):
                ports = [ports]
            mirror_interface_conf['mirror'] = ports
        # Don't delete DP level config when setting mirror list to empty,
        # as that could cause an unnecessary cold start.
        elif 'mirror' in mirror_interface_conf:
            del mirror_interface_conf['mirror']
        self.set_port_conf(dp, mirror_port, mirror_interface_conf)

    def mirror_port(self, dp, mirror_port, port):
        mirror_interface_conf, ports = self._get_mirrored_ports(
            dp, mirror_port)
        if ports is None:
            ports = []
        ports = set(ports)
        ports.add(port)
        self._set_mirror_config(dp, mirror_port, mirror_interface_conf, ports)

    def unmirror_port(self, dp, mirror_port, port):
        mirror_interface_conf, ports = self._get_mirrored_ports(
            dp, mirror_port)
        if ports is None:
            ports = []
        ports = set(ports)
        if port in ports:
            ports.remove(port)
            self._set_mirror_config(
                dp, mirror_port, mirror_interface_conf, ports)

    def clear_mirror_port(self, dp, mirror_port):
        mirror_interface_conf, _ = self._get_mirrored_ports(dp, mirror_port)
        self._set_mirror_config(dp, mirror_port, mirror_interface_conf)


class FaucetRemoteConfGetSetter(FaucetConfGetSetter):

    def __init__(self, client_key=None, client_cert=None,
                 ca_cert=None, server_addr=None):
        self.client = FaucetConfRpcClient(
            client_key=client_key, client_cert=client_cert,
            ca_cert=ca_cert, server_addr=server_addr)

    @staticmethod
    def config_file_path(config_file):
        if config_file:
            return os.path.basename(config_file)
        return config_file

    def read_faucet_conf(self, config_file):
        self.faucet_conf = self.client.get_config_file(
            config_filename=self.config_file_path(config_file))
        if self.faucet_conf is None:
            logging.error('Faucet config is empty, exiting.')
            sys.exit(1)
        return self.faucet_conf

    def write_faucet_conf(self, config_file=None, faucet_conf=None, merge=False):
        if not config_file:
            config_file = self.DEFAULT_CONFIG_FILE
        if faucet_conf is None:
            faucet_conf = self.faucet_conf
        return self.client.set_config_file(
            self.faucet_conf,
            config_filename=self.config_file_path(config_file),
            merge=merge)

    def get_dps(self):
        self.read_faucet_conf(config_file=None)
        return self.faucet_conf.get('dps', {})

    def set_port_conf(self, dp, port, port_conf):
        return self.client.set_dp_interfaces(
            [(dp, {port: yaml.dump(port_conf)})])

    def update_switch_conf(self, dp, switch_conf):
        return self.write_faucet_conf(
            faucet_conf={'dps': {dp: switch_conf}}, merge=True)

    def mirror_port(self, dp, mirror_port, port):  # pragma: no cover
        self.client.add_port_mirror(dp, port, mirror_port)

    def unmirror_port(self, dp, mirror_port, port):  # pragma: no cover
        self.client.remove_port_mirror(dp, port, mirror_port)

    def clear_mirror_port(self, dp, mirror_port):  # pragma: no cover
        self.client.clear_port_mirror(dp, mirror_port)



# -*- coding: utf-8 -*-
"""
Created on 4 March 2020
@author: Charlie Lewis
"""
import os
import tempfile

import yaml

from poseidon_core.helpers.exception_decor import exception


def represent_none(dumper, _):
    return dumper.represent_scalar('tag:yaml.org,2002:null', '')


@exception
def get_config_file(config_file):
    # TODO check for other files
    if not config_file:
        # default to FAUCET default
        config_file = '/etc/faucet/faucet.yaml'
    return config_file


def parse_rules(config_file):
    config_file = get_config_file(config_file)
    obj_doc = yaml_in(config_file)
    return obj_doc


@exception
def yaml_in(config_file):
    try:
        with open(config_file, 'r') as stream:
            return yaml.safe_load(stream)
    except Exception as e:  # pragma: no cover
        return False


@exception
def yaml_out(config_file, obj_doc):
    stream = tempfile.NamedTemporaryFile(
        prefix=os.path.basename(config_file),
        dir=os.path.dirname(config_file),
        mode='w',
        delete=False)
    yaml.add_representer(type(None), represent_none)
    yaml.dump(obj_doc, stream, default_flow_style=False)
    stream.close()
    os.replace(stream.name, config_file)
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import queue
import random
import re
import sys
import threading
import time
from collections import defaultdict
from functools import partial

import requests
import schedule

from poseidon_core.helpers.actions import Actions
from poseidon_core.helpers.config import Config
from poseidon_core.helpers.prometheus import Prometheus
from poseidon_core.sdnconnect import SDNConnect

requests.packages.urllib3.disable_warnings()


class Monitor:

    def __init__(self, logger, controller=None):
        self.logger = logger
        self.m_queue = queue.Queue()
        self.job_queue = queue.Queue()
        self.rabbits = []
        self.running = True

        # get config options
        if controller is None:
            self.controller = Config().get_config()
        else:
            self.controller = controller

        # setup prometheus
        self.prom = Prometheus()
        try:
            self.prom.initialize_metrics()
        except Exception as e:  # pragma: no cover
            self.logger.debug(
                'Prometheus metrics are already initialized: {0}'.format(str(e)))
        Prometheus.start()

        # initialize sdnconnect
        self.s = SDNConnect(self.controller, self.logger)
        self.s.default_endpoints()
        self.update_endpoint_metadata()

        # timer class to call things periodically in own thread
        self.schedule = schedule
        self.schedule.every(self.controller['scan_frequency']).seconds.do(
            self.schedule_job_update_metrics)
        self.schedule.every(self.controller['reinvestigation_frequency']).seconds.do(
            self.schedule_job_reinvestigation_timeout)

        # schedule all threads
        self.schedule_thread = threading.Thread(
            target=partial(
                self.schedule_thread_worker,
                scheduler=self.schedule),
            name='st_worker')

    def schedule_thread_worker(self, scheduler=schedule):
        ''' schedule thread, takes care of running processes in the future '''
        self.logger.debug('Starting thread_worker')
        while self.running:
            sys.stdout.flush()
            scheduler.run_pending()
            time.sleep(1)

    def ignore_rabbit(self, routing_key, body):
        ''' drop ignored messages. '''
        if routing_key == self.controller['FA_RABBIT_ROUTING_KEY']:
            if self.s and self.s.sdnc:
                if self.s.sdnc.ignore_event(body):
                    return True
        return False

    def rabbit_callback(self, ch, method, _properties, body, q=None):
        ''' callback, places rabbit data into internal queue'''
        body = json.loads(body)
        self.logger.debug('got a message: {0}:{1} (qsize {2})'.format(
            method.routing_key, body, q.qsize()))
        if q is not None:
            self.update_routing_key_time(method.routing_key)
            if not self.ignore_rabbit(method.routing_key, body):
                q.put((method.routing_key, body))
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def get_hosts(self):
        # TODO consolidate with update_endpoint_metadata
        hosts = []
        for hash_id, endpoint in self.s.endpoints.items():
            roles, _, _ = endpoint.get_roles_confidences_pcap_labels()
            role = roles[0]
            ipv4_os = endpoint.get_ipv4_os()
            host = {
                'mac': endpoint.endpoint_data['mac'],
                'id': hash_id,
                'role': role,
                'ipv4_os': ipv4_os,
                'state': endpoint.state,
                'tenant': endpoint.endpoint_data['tenant'],
                'port': endpoint.endpoint_data['port'],
                'segment': endpoint.endpoint_data['segment'],
                'ipv4': endpoint.endpoint_data['ipv4']}
            hosts.append(host)
        return hosts

    def job_update_metrics(self):
        self.logger.debug('updating metrics')
        try:
            hosts = self.get_hosts()
            self.prom.update_metrics(hosts)
        except (requests.exceptions.ConnectionError, Exception) as e:  # pragma: no cover
            self.logger.error(
                'Unable to get current state and send it to Prometheus because: {0}'.format(str(e)))
        return 0

    def job_recoprocess(self):
        if not self.s.sdnc:
            for endpoint in self.s.not_copro_ignored_endpoints():
                if endpoint.copro_state != 'copro_nominal':
                    endpoint.copro_nominal()  # pytype: disable=attribute-error
            return 0
        events = 0
        for endpoint in self.s.not_copro_ignored_endpoints('copro_coprocessing'):
            if endpoint.copro_state_timeout(2*self.controller['coprocessing_frequency']):
                self.logger.debug(
                    'timing out: {0} and setting to unknown'.format(endpoint.name))
                self.s.uncoprocess_endpoint(endpoint)
                endpoint.copro_unknown()  # pytype: disable=attribute-error
                events += 1
        return events

    def job_reinvestigation_timeout(self):
        ''' put endpoints into the reinvestigation state if possible, and timeout investigations '''
        if not self.s.sdnc:
            for endpoint in self.s.not_ignored_endpoints():
                if endpoint.state != 'known':
                    endpoint.known()  # pytype: disable=attribute-error
            return 0
        events = 0
        timeout = 2*self.controller['reinvestigation_frequency']
        for endpoint in self.s.not_ignored_endpoints():
            if endpoint.observed_timeout(timeout):
                self.logger.info('observation timing out: {0}'.format(endpoint.name))
                endpoint.force_unknown()
                events += 1
            elif endpoint.mirror_active() and endpoint.state_timeout(timeout):
                self.logger.info('mirror timing out: {0}'.format(endpoint.name))
                self.s.unmirror_endpoint(endpoint)
                events += 1
        budget = self.s.investigation_budget()
        candidates = self.s.not_ignored_endpoints('queued')
        if not candidates:
            candidates = self.s.not_ignored_endpoints('known')
        return events + self._schedule_queued_work(
            candidates, budget, 'reinvestigate', self.s.mirror_endpoint, shuffle=True)

    def queue_job(self, job):
        if self.job_queue.qsize() < 2:
            self.job_queue.put(job)

    def schedule_job_update_metrics(self):
        self.queue_job(self.job_update_metrics)

    def schedule_job_reinvestigation_timeout(self):
        self.queue_job(self.job_reinvestigation_timeout)

    def update_routing_key_time(self, routing_key):
        if self.prom:
            self.prom.prom_metrics['last_rabbitmq_routing_key_time'].labels(
                routing_key=routing_key).set(time.time())

    def update_endpoint_metadata(self):
        update_time = time.time()
        for hash_id, endpoint in self.s.endpoints.items():
            ipv4 = endpoint.endpoint_data['ipv4']
            ipv6 = endpoint.endpoint_data['ipv6']
            ipv4_subnet = endpoint.endpoint_data['ipv4_subnet']
            ipv6_subnet = endpoint.endpoint_data['ipv6_subnet']
            ipv4_rdns = endpoint.endpoint_data['ipv4_rdns']
            ipv6_rdns = endpoint.endpoint_data['ipv6_rdns']
            port = endpoint.endpoint_data['port']
            tenant = endpoint.endpoint_data['tenant']
            segment = endpoint.endpoint_data['segment']
            ether_vendor = endpoint.endpoint_data['ether_vendor']
            controller = endpoint.endpoint_data['controller']
            controller_type = endpoint.endpoint_data['controller_type']
            roles, confidences, pcap_labels = endpoint.get_roles_confidences_pcap_labels()
            top_role, second_role, third_role = roles
            top_conf, second_conf, third_conf = confidences
            ipv4_os = endpoint.get_ipv4_os()

            def set_prom(var, val, **prom_labels):
                prom_labels.update({
                    'mac': endpoint.endpoint_data['mac'],
                    'name': endpoint.endpoint_data['name'],
                    'hash_id': hash_id,
                })
                try:
                    self.prom.prom_metrics[var].labels(**prom_labels).set(val)
                except ValueError:
                    pass

            def set_prom_role(var, val, role):
                set_prom(
                    var,
                    val,
                    role=role,
                    ipv4_os=ipv4_os,
                    ipv4_address=ipv4,
                    ipv6_address=ipv6,
                    pcap_labels=pcap_labels)

            def update_prom(var, **prom_labels):
                prom_labels.update({
                    'tenant': tenant,
                    'segment': segment,
                    'ether_vendor': ether_vendor,
                    'port': port,
                })
                set_prom(var, update_time, **prom_labels)

            set_prom_role(
                'endpoint_role_confidence_top',
                top_conf,
                top_role)
            set_prom_role(
                'endpoint_role_confidence_second',
                second_conf,
                second_role)
            set_prom_role(
                'endpoint_role_confidence_third',
                third_conf,
                third_role)
            update_prom(
                'endpoints',
                controller_type=controller_type,
                controller=controller)
            update_prom(
                'endpoint_state',
                state=endpoint.state)
            update_prom(
                'endpoint_os',
                ipv4_os=ipv4_os)
            update_prom(
                'endpoint_role',
                top_role=top_role)
            update_prom(
                'endpoint_ip',
                ipv4_subnet=ipv4_subnet,
                ipv6_subnet=ipv6_subnet,
                ipv4_rdns=ipv4_rdns,
                ipv6_rdns=ipv6_rdns,
                ipv4_address=ipv4,
                ipv6_address=ipv6)
            update_prom(
                'endpoint_metadata',
                prev_state=endpoint.p_prev_state,
                next_state=endpoint.p_next_state,
                acls=endpoint.acl_data,
                ignore=str(endpoint.ignore),
                ipv4_subnet=ipv4_subnet,
                ipv6_subnet=ipv6_subnet,
                ipv4_rdns=ipv4_rdns,
                ipv6_rdns=ipv6_rdns,
                controller_type=controller_type,
                controller=controller,
                state=endpoint.state,
                top_role=top_role,
                ipv4_os=ipv4_os,
                ipv4_address=ipv4,
                ipv6_address=ipv6)

    def merge_metadata(self, new_metadata):
        updated = set()
        metadata_types = {
            'mac_addresses': self.s.endpoints_by_mac,
            'ipv4_addresses': self.s.endpoints_by_ip,
            'ipv6_addresses': self.s.endpoints_by_ip,
        }
        for metadata_type, metadata_lookup in metadata_types.items():
            type_new_metadata = new_metadata.get(metadata_type, {})
            for key, data in type_new_metadata.items():
                endpoints = metadata_lookup(key)
                if endpoints:
                    endpoint = endpoints[0]
                    if metadata_type not in endpoint.metadata:
                        endpoint.metadata[metadata_type] = defaultdict(dict)
                    endpoint.metadata[metadata_type][key].update(data)
                    updated.add(endpoint)
        return updated

    def format_rabbit_message(self, item, faucet_event, remove_list):
        '''
        read a message off the rabbit_q
        the message should be item = (routing_key,msg)
        '''
        routing_key, my_obj = item
        self.logger.debug(
            'routing_key: {0} rabbit_message: {1}'.format(routing_key, my_obj))

        def handler_algos_decider(my_obj):
            self.logger.debug('decider value:{0}'.format(my_obj))
            data = my_obj.get('data', None)
            results = my_obj.get('results', {})
            tool = results.get('tool', None)
            if isinstance(data, dict) and data:
                new_metadata = data
                if tool == 'p0f':
                    ip_metadata = {}
                    for ip, ip_data in data.items():
                        if ip_data and ip_data.get('full_os', None):
                            ip_metadata[ip] = ip_data
                    new_metadata = {'ipv4_addresses': ip_metadata}
                elif tool == 'networkml':
                    mac_metadata = {}
                    for name, message in data.items():
                        if name == 'pcap':
                            continue
                        if message.get('valid', False):
                            source_mac = message.get('source_mac', None)
                            if source_mac:
                                mac_metadata[source_mac] = message
                    new_metadata = {'mac_addresses': mac_metadata}
                # Generic handler for future tools.
                updated = self.merge_metadata(new_metadata)
                if updated:
                    for endpoint in updated:
                        if endpoint.mirror_active():
                            self.s.unmirror_endpoint(endpoint)
                    return data
            return {}

        def handler_action_ignore(my_obj):
            for name in my_obj:
                endpoint = self.s.endpoints.get(name, None)
                if endpoint:
                    endpoint.ignore = True
            return {}

        def handler_action_clear_ignored(my_obj):
            for name in my_obj:
                endpoint = self.s.endpoints.get(name, None)
                if endpoint:
                    endpoint.ignore = False
            return {}

        def handler_action_change(my_obj):
            for name, state in my_obj:
                endpoint = self.s.endpoints.get(name, None)
                if endpoint:
                    try:
                        if endpoint.mirror_active():
                            self.s.unmirror_endpoint(endpoint)
                        endpoint.machine_trigger(state)  # pytype: disable=attribute-error
                        endpoint.p_next_state = None
                        if endpoint.mirror_active():
                            self.s.mirror_endpoint(endpoint)
                            self.prom.prom_metrics['ncapture_count'].inc()
                    except Exception as e:  # pragma: no cover
                        self.logger.error(
                            'Unable to change endpoint {0} because: {1}'.format(endpoint.name, str(e)))
            return {}

        def handler_action_update_acls(my_obj):
            for ip in my_obj:
                rules = my_obj[ip]
                endpoints = self.s.endpoints_by_ip(ip)
                if endpoints:
                    endpoint = endpoints[0]
                    try:
                        status = Actions(
                            endpoint, self.s.sdnc).update_acls(
                                rules_file=self.controller['RULES_FILE'], endpoints=endpoints, force_apply_rules=rules)
                        if not status:
                            self.logger.warning(
                                'Unable to apply rules: {0} to endpoint: {1}'.format(rules, endpoint.name))
                    except Exception as e:
                        self.logger.error(
                            'Unable to apply rules: {0} to endpoint: {1} because {2}'.format(rules, endpoint.name, str(e)))
            return {}

        def handler_action_remove(my_obj):
            remove_list.extend([name for name in my_obj])
            return {}

        def handler_action_remove_ignored(_my_obj):
            remove_list.extend([
                endpoint.name for endpoint in self.s.endpoints.values()
                if endpoint.ignore])
            return {}

        def handler_faucet_event(my_obj):
            if self.s and self.s.sdnc:
                faucet_event.append(my_obj)
                return my_obj
            return {}

        handlers = {
            'poseidon.algos.decider': handler_algos_decider,
            'poseidon.action.ignore': handler_action_ignore,
            'poseidon.action.clear.ignored': handler_action_clear_ignored,
            'poseidon.action.change': handler_action_change,
            'poseidon.action.update_acls': handler_action_update_acls,
            'poseidon.action.remove': handler_action_remove,
            'poseidon.action.remove.ignored': handler_action_remove_ignored,
            self.controller['FA_RABBIT_ROUTING_KEY']: handler_faucet_event,
        }

        handler = handlers.get(routing_key, None)
        if handler is not None:
            ret_val = handler(my_obj)
            return ret_val, True

        self.logger.error(
            'no handler for routing_key {0}'.format(routing_key))
        return {}, False

    def _schedule_queued_work(self, queued_endpoints, budget, endpoint_state, endpoint_work, shuffle=False):
        events = 0
        if self.s.sdnc:
            if shuffle:
                random.shuffle(queued_endpoints)
            for endpoint in queued_endpoints[:budget]:
                getattr(endpoint, endpoint_state)()
                endpoint_work(endpoint)
                if endpoint_state in ['trigger_next', 'reinvestigate']:
                    self.prom.prom_metrics['ncapture_count'].inc()
                events += 1
        return events

    def schedule_mirroring(self):
        for endpoint in self.s.not_ignored_endpoints('unknown'):
            endpoint.queue_next('mirror')
        budget = self.s.investigation_budget()
        queued_endpoints = [
            endpoint for endpoint in self.s.not_ignored_endpoints('queued')
            if endpoint.mirror_requested()]
        queued_endpoints = sorted(queued_endpoints, key=lambda x: x.state_time)
        self.logger.debug('investigations {0}, budget {1}, queued {2}'.format(
            str(self.s.investigations), str(budget), str(len(queued_endpoints))))
        return self._schedule_queued_work(queued_endpoints, budget, 'trigger_next', self.s.mirror_endpoint)

    def schedule_coprocessing(self):
        for endpoint in self.s.not_copro_ignored_endpoints('copro_unknown'):
            endpoint.copro_queue_next('copro_coprocess')
        budget = self.s.coprocessing_budget()
        queued_endpoints = self.s.not_copro_ignored_endpoints('copro_queued')
        queued_endpoints = sorted(queued_endpoints, key=lambda x: x.copro_state_time)
        self.logger.debug('coprocessing {0}, budget {1}, queued {2}'.format(
            str(self.s.coprocessing), str(budget), str(len(queued_endpoints))))
        return self._schedule_queued_work(queued_endpoints, budget, 'copro_trigger_next', self.s.coprocess_endpoint)

    def monitor_callable(self, monitored_callable):
        method_name = str(monitored_callable)
        method_re = re.compile(r'.+bound method (\S+).+')
        method_match = method_re.match(method_name)
        if method_match:
            method_name = method_match.group(1)
        with self.prom.prom_metrics['monitor_runtime_secs'].labels(method=method_name).time():
            return monitored_callable()

    def handle_rabbit(self):
        events = 0
        faucet_event = []
        remove_list = []
        while True:
            found_work, rabbit_msg = self.monitor_callable(partial(self.get_q_item, self.m_queue))
            if not found_work:
                break
            events += 1
            # faucet_event and remove_list get updated as references because partial()
            self.monitor_callable(partial(self.format_rabbit_message, rabbit_msg, faucet_event, remove_list))
        return (events, faucet_event, remove_list)

    def process(self):
        while self.running:
            events, faucet_event, remove_list = self.monitor_callable(self.handle_rabbit)
            if remove_list:
                for endpoint_name in remove_list:
                    if endpoint_name in self.s.endpoints:
                        del self.s.endpoints[endpoint_name]
            if faucet_event:
                self.monitor_callable(partial(self.s.check_endpoints, faucet_event))
            events += self.monitor_callable(self.schedule_mirroring)
            found_work, schedule_func = self.monitor_callable(partial(self.get_q_item, self.job_queue))
            if found_work and callable(schedule_func):
                events += self.monitor_callable(schedule_func)
            if events:
                self.monitor_callable(self.update_endpoint_metadata)
            time.sleep(1)

    @staticmethod
    def get_q_item(q):
        '''
        attempt to get a work item from the queue
        m_queue -> (routing_key, body)
        a read from get_q_item should be of the form
        (boolean,(routing_key, body))
        '''
        try:
            item = q.get_nowait()
            q.task_done()
            return (True, item)
        except queue.Empty:  # pragma: no cover
            pass

        return (False, None)
